--****************************************************************************************************
-- Client Lua Script for AchievementTracker
-- Copyright (c) NCsoft. All rights reserved
-- Created by SpielePhilosoph - http://www.curse.com/users/SpielePhilosoph/projects
-- Last modified: 201407015
-- Version: 0.7
--****************************************************************************************************

--****************************************************************************************************
-- List of used libs, that need for this addon.
--****************************************************************************************************
require "AchievementsLib"
require "Apollo"
require "Window"

--****************************************************************************************************
-- AchievementTracker Module Definition
--****************************************************************************************************
local AchievementTracker = {}

--****************************************************************************************************
-- AchievementTracker Variables
--****************************************************************************************************
-- Infos for debugging the addon.
local ShowDebugInfos = false
-- Functions for debug use:
-- Event_FireGenericEvent("SendVarToRover", "itemName", item) -- Send a item to rover for detail validation.
-- Print(debug.traceback()) -- Prints the current stacktrace.

local ConfigMinVersion = 2 -- Minumum version for config data to load it.

-- Achievement Tacker Modes.
local ModeAllId = 1
local ModeCustomListId = 3
local ModeFilterId = 2
local ModeRecentlyChangedId = 4
local AchievementTrackerMode =
{
 {
  Id = ModeAllId,
  Name = "All",
  Tooltip = "Shows all achievements in the list.",
 },
 {
  Id = ModeFilterId,
  Name = "Filter",
  Tooltip = "Shows achievements that match filter selection.",
 },
 {
  Id = ModeCustomListId,
  Name = "Custom List",
  Tooltip = "Shows achievements that are in the custom list.",
 },
 {
  Id = ModeRecentlyChangedId,
  Name = "Recently Changed",
  Tooltip = "Shows achievements that were updated last time.",
 },
}

-- Strings to used in the addon.
local NamePrefixGuild = "(G)"

-- Icon mapping to achievement categories.
local AchievementIconCategoryMapping =
{
 [1] = "IconSprites:Icon_Achievement_Achievement_Exploration", -- Erkundung
 [193] = "IconSprites:Icon_Achievement_Achievement_Exploration", -- Immersternwald
 [194] = "IconSprites:Icon_Achievement_Achievement_Exploration", -- Nrdliche Wildnis
 [195] = "IconSprites:Icon_Achievement_Achievement_Exploration", -- Deradune
 [197] = "IconSprites:Icon_Achievement_Achievement_Exploration", -- Algofels
 [198] = "IconSprites:Icon_Achievement_Achievement_Exploration", -- Celestia
 [200] = "IconSprites:Icon_Achievement_Achievement_Exploration", -- Orkanas
 [201] = "IconSprites:Icon_Achievement_Achievement_Exploration", -- Weital
 [202] = "IconSprites:Icon_Achievement_Achievement_Exploration", -- Fernseits
 [203] = "IconSprites:Icon_Achievement_Achievement_Exploration", -- Wildlauf
 [204] = "IconSprites:Icon_Achievement_Achievement_Exploration", -- Jochgrab
 [205] = "IconSprites:Icon_Achievement_Achievement_Exploration", -- Gramkammer
 [207] = "IconSprites:Icon_Achievement_Achievement_Exploration", -- Thayd
 [3] = "IconSprites:Icon_Achievement_Achievement_Quest", -- Quest
 [178] = "IconSprites:Icon_Achievement_Achievement_Quest", -- Immersternwald
 [179] = "IconSprites:Icon_Achievement_Achievement_Quest", -- Nrdliche Wildnis
 [182] = "IconSprites:Icon_Achievement_Achievement_Quest", -- Algofels
 [183] = "IconSprites:Icon_Achievement_Achievement_Quest", -- Celestia
 [185] = "IconSprites:Icon_Achievement_Achievement_Quest", -- Orkanas
 [186] = "IconSprites:Icon_Achievement_Achievement_Quest", -- Weital
 [187] = "IconSprites:Icon_Achievement_Achievement_Quest", -- Fernseits
 [188] = "IconSprites:Icon_Achievement_Achievement_Quest", -- Wildlauf
 [189] = "IconSprites:Icon_Achievement_Achievement_Quest", -- Jochgrab
 [190] = "IconSprites:Icon_Achievement_Achievement_Quest", -- Gramkammer
 [4] = "IconSprites:Icon_Achievement_Achievement_Combat", -- Kopfjagd
 [210] = "IconSprites:Icon_Achievement_Achievement_Combat", -- Immersternwald
 [211] = "IconSprites:Icon_Achievement_Achievement_Combat", -- Nrdliche Wildnis
 [214] = "IconSprites:Icon_Achievement_Achievement_Combat", -- Algofels
 [215] = "IconSprites:Icon_Achievement_Achievement_Combat", -- Celestia
 [217] = "IconSprites:Icon_Achievement_Achievement_Combat", -- Orkanas
 [218] = "IconSprites:Icon_Achievement_Achievement_Combat", -- Weital
 [219] = "IconSprites:Icon_Achievement_Achievement_Combat", -- Fernseits
 [220] = "IconSprites:Icon_Achievement_Achievement_Combat", -- Wildlauf
 [221] = "IconSprites:Icon_Achievement_Achievement_Combat", -- Jochgrab
 [222] = "IconSprites:Icon_Achievement_Achievement_Combat", -- Gramkammer
 [282] = "IconSprites:Icon_Achievement_Achievement_Combat", -- Isolationsanlage R-12
 [283] = "IconSprites:Icon_Achievement_Achievement_Combat", -- Blutrotes Hinterland
 [295] = "IconSprites:Icon_Achievement_Achievement_Combat", -- Nrdliches dland
 [7] = "IconSprites:Icon_Achievement_Achievement_PvP", -- PvP
 [69] = "IconSprites:Icon_Achievement_Achievement_PvP", -- Arenen
 [153] = "IconSprites:Icon_Achievement_Achievement_PvP", -- 2v2-Arena
 [154] = "IconSprites:Icon_Achievement_Achievement_PvP", -- 3v3-Arena
 [155] = "IconSprites:Icon_Achievement_Achievement_PvP", -- 5v5-Arena
 [70] = "IconSprites:Icon_Achievement_Achievement_PvP", -- Schlachtfelder
 [149] = "IconSprites:Icon_Achievement_Achievement_PvP", -- Walatiki-Tempel
 [150] = "IconSprites:Icon_Achievement_Achievement_PvP", -- Hallen der Blutgeschworenen
 [71] = "IconSprites:Icon_Achievement_Achievement_PvP", -- Kriegsbasen
 [156] = "IconSprites:Icon_Achievement_Achievement_PvP", -- Allgemein
 [9] = "IconSprites:Icon_Achievement_Achievement_Path_Explorer", -- Pfade
 [11] = "IconSprites:Icon_Achievement_Achievement_Reputation", -- Ansehen
 [246] = "IconSprites:Icon_Achievement_Achievement_Reputation", -- Algofels
 [247] = "IconSprites:Icon_Achievement_Achievement_Reputation", -- Celestia
 [249] = "IconSprites:Icon_Achievement_Achievement_Reputation", -- Orkanas
 [250] = "IconSprites:Icon_Achievement_Achievement_Reputation", -- Weital
 [251] = "IconSprites:Icon_Achievement_Achievement_Reputation", -- Fernseits
 [252] = "IconSprites:Icon_Achievement_Achievement_Reputation", -- Wildlauf
 [253] = "IconSprites:Icon_Achievement_Achievement_Reputation", -- Jochgrab
 [254] = "IconSprites:Icon_Achievement_Achievement_Reputation", -- Gramkammer
 [12] = "IconSprites:Icon_Achievement_Achievement_GenericAchievement", -- Allgemein
 [65] = "IconSprites:Icon_Achievement_Achievement_Dungeon", -- Dungeons
 [66] = "IconSprites:Icon_Achievement_Achievement_Dungeon", -- Feuerschdel
 [67] = "IconSprites:Icon_Achievement_Achievement_Dungeon", -- Modus: Normal
 [68] = "IconSprites:Icon_Achievement_Achievement_Dungeon", -- Modus: Erfahren
 [166] = "IconSprites:Icon_Achievement_Achievement_Dungeon", -- Sturmklaue
 [167] = "IconSprites:Icon_Achievement_Achievement_Dungeon", -- Modus: Normal
 [168] = "IconSprites:Icon_Achievement_Achievement_Dungeon", -- Modus: Erfahren
 [169] = "IconSprites:Icon_Achievement_Achievement_Dungeon", -- Zuflucht der Schwertmaid
 [170] = "IconSprites:Icon_Achievement_Achievement_Dungeon", -- Modus: Normal
 [171] = "IconSprites:Icon_Achievement_Achievement_Dungeon", -- Modus: Erfahren
 [172] = "IconSprites:Icon_Achievement_Achievement_Dungeon", -- Kel Voreth
 [173] = "IconSprites:Icon_Achievement_Achievement_Dungeon", -- Modus: Normal
 [174] = "IconSprites:Icon_Achievement_Achievement_Dungeon", -- Modus: Erfahren
 [78] = "IconSprites:Icon_Achievement_Achievement_Adventures", -- Abenteuer
 [79] = "IconSprites:Icon_Achievement_Achievement_Adventures", -- Treck durch Jochgrab
 [80] = "IconSprites:Icon_Achievement_Achievement_Adventures", -- Modus: Normal
 [81] = "IconSprites:Icon_Achievement_Achievement_Adventures", -- Modus: Erfahren
 [82] = "IconSprites:Icon_Achievement_Achievement_Adventures", -- Banden von Weital
 [83] = "IconSprites:Icon_Achievement_Achievement_Adventures", -- Modus: Normal
 [84] = "IconSprites:Icon_Achievement_Achievement_Adventures", -- Modus: Erfahren
 [85] = "IconSprites:Icon_Achievement_Achievement_Adventures", -- Die Belagerung der Sturmzuflucht
 [86] = "IconSprites:Icon_Achievement_Achievement_Adventures", -- Modus: Normal
 [87] = "IconSprites:Icon_Achievement_Achievement_Adventures", -- Modus: Erfahren
 [88] = "IconSprites:Icon_Achievement_Achievement_Adventures", -- Krieg der Wildnis
 [89] = "IconSprites:Icon_Achievement_Achievement_Adventures", -- Modus: Normal
 [90] = "IconSprites:Icon_Achievement_Achievement_Adventures", -- Modus: Erfahren
 [91] = "IconSprites:Icon_Achievement_Achievement_Adventures", -- Viel Lrm um Nichts
 [92] = "IconSprites:Icon_Achievement_Achievement_Adventures", -- Modus: Normal
 [94] = "IconSprites:Icon_Achievement_Achievement_Adventures", -- Der Kronfelder Aufstand
 [95] = "IconSprites:Icon_Achievement_Achievement_Adventures", -- Modus: Normal
 [161] = "IconSprites:Icon_Achievement_Achievement_Challenges", -- Herausforderungen
 [261] = "IconSprites:Icon_Achievement_Achievement_Challenges", -- Algofels
 [262] = "IconSprites:Icon_Achievement_Achievement_Challenges", -- Celestia
 [264] = "IconSprites:Icon_Achievement_Achievement_Challenges", -- Orkanas
 [265] = "IconSprites:Icon_Achievement_Achievement_Challenges", -- Weital
 [266] = "IconSprites:Icon_Achievement_Achievement_Challenges", -- Fernseits
 [267] = "IconSprites:Icon_Achievement_Achievement_Challenges", -- Wildlauf
 [268] = "IconSprites:Icon_Achievement_Achievement_Challenges", -- Jochgrab
 [269] = "IconSprites:Icon_Achievement_Achievement_Challenges", -- Gramkammer
 [273] = "IconSprites:Icon_Achievement_Achievement_Challenges", -- Thayd
 [162] = "IconSprites:Icon_Achievement_Achievement_Shiphand", -- Schiffsjunge
 [274] = "IconSprites:Icon_Achievement_Achievement_Shiphand", -- Auenposten M-13
 [275] = "IconSprites:Icon_Achievement_Achievement_Shiphand", -- Befall
 [276] = "IconSprites:Icon_Achievement_Achievement_Shiphand", -- Rasende Logik
 [277] = "IconSprites:Icon_Achievement_Achievement_Shiphand", -- Weltraumkoller
 [278] = "IconSprites:Icon_Achievement_Achievement_Shiphand", -- Verschwunden im Weltraum
 [279] = "IconSprites:Icon_Achievement_Achievement_Shiphand", -- Spierutenlauf
 [163] = "IconSprites:Icon_Achievement_Achievement_WorldEvent", -- ffentliche Events
 [230] = "IconSprites:Icon_Achievement_Achievement_WorldEvent", -- Algofels
 [231] = "IconSprites:Icon_Achievement_Achievement_WorldEvent", -- Celestia
 [233] = "IconSprites:Icon_Achievement_Achievement_WorldEvent", -- Orkanas
 [234] = "IconSprites:Icon_Achievement_Achievement_WorldEvent", -- Weital
 [235] = "IconSprites:Icon_Achievement_Achievement_WorldEvent", -- Fernseits
 [236] = "IconSprites:Icon_Achievement_Achievement_WorldEvent", -- Wildlauf
 [237] = "IconSprites:Icon_Achievement_Achievement_WorldEvent", -- Jochgrab
 [238] = "IconSprites:Icon_Achievement_Achievement_WorldEvent", -- Gramkammer
 [164] = "IconSprites:Icon_Achievement_Achievement_Raid", -- Schlachtzge
 [165] = "IconSprites:Icon_Achievement_Achievement_Raid", -- Datenzone
 [239] = "IconSprites:Icon_Achievement_Achievement_Raid", -- Genarchiv
 [297] = "IconSprites:Icon_Achievement_Achievement_Raid", -- Sicherheitsschlssel
 -- The carbine addon used IconSprites:Icon_Achievement_Achievement_WorldEvent.
 [175] = "IconSprites:Icon_Achievement_Achievement_GalacticArchives", -- Geschichte der Welt
 [281] = "IconSprites:Icon_Achievement_Achievement_Social", -- Sozial
 [284] = "IconSprites:Icon_Achievement_Achievement_ServerWide_General", -- Galaktischer Teufelskerl
}

--****************************************************************************************************
-- Initialization
--****************************************************************************************************
------------------------------------------------------------------------------------------------------
-- AchievementTracker new
------------------------------------------------------------------------------------------------------
function AchievementTracker:new(o)
 o = o or {}
 setmetatable(o, self)
 self.__index = self 

 o.AchievmentItemList = {} -- List of all current listed achievement items to easy update there informations.
 o.Config = {} -- Property to saved and load config data.
 o.Config.CustomList = {} -- Property to save the id of achievements that are shown in the custom list.
 o.Config.AchievementTrackerMode = ModeRecentlyChangedId -- Default mode is filter.
 o.Config.FilterCategory = {} -- List of categories for filter.
 o.Config.FilterZone = {} -- List of zones for filter.
 o.Config.IsVisible = false -- Per default the tracker window is not shown.
 o.Config.NegativeCategoryFilter = false -- Per default the category is used as positive filter.
 o.Config.PositionMain = -- Default position for AchievementTrackerForm.
 {
  Bottom = 500,
  Left = 0,
  Right = 300,
  Top = 0,
 }
 o.Config.PositionOptions = -- Default position for AchievementTrackerOptionsForm.
 {
  Bottom = 500,
  Left = 110,
  Right = 600,
  Top = 0,
 }
 o.Config.RecentlyChangedList = {} -- Property to save the id of achievements that are recently updated.
 o.Config.ShowCompletedAchievements = true -- Per default completed achievements are shown.
 o.Config.ShowGuildAchievements = false -- Per default the guild achievements not shown.
 o.Config.Version = 2 -- Version of config.
 o.ZoneInfo = {} -- List for achievement zones if they have achievements or not.
 return o
end

------------------------------------------------------------------------------------------------------
-- AchievementTracker Init
------------------------------------------------------------------------------------------------------
function AchievementTracker:Init()
 local bHasConfigureFunction = true
 local strConfigureButtonText = "Achv. Tracker"
 local tDependencies = {}
 if ShowDebugInfos == true then
  -- For debug make sure, that Chat and Rover ist loaded an can be used.
  table.insert(tDependencies, "ChatLog")
  table.insert(tDependencies, "Rover")
 end
 Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end

--****************************************************************************************************
-- Eventfunctions for AchievementTracker
--****************************************************************************************************
------------------------------------------------------------------------------------------------------
-- AchievementTracker OnAchievementTrackerCall
-- on SlashCommand "/achvtrack"
-- Shows the main form of the achievement tracker, register for the AchievementUpdated event and
-- start update timer.
------------------------------------------------------------------------------------------------------
function AchievementTracker:OnAchievementTrackerCall()
 if self.MainForm:IsShown() == true then
  self:MainHide()
 else
  self:MainShow()
 end
end

------------------------------------------------------------------------------------------------------
-- AchievementTracker OnAchievementUpdated
-- This function update one achievment in the list.
------------------------------------------------------------------------------------------------------
function AchievementTracker:OnAchievementUpdated(updatedAchievement)
 if updatedAchievement ~= nil then
  local achievementId = updatedAchievement:GetId()
  local updateSingelItem = true
  -- Update achievement points.
  self:UpdateTitel()
  -- Add achievement to the list and rebuild in if the items was new and mode is RecentlyChanged.
  if self.Config ~= nil and
     self.Config.RecentlyChangedList[achievementId] == nil then
   self.Config.RecentlyChangedList[achievementId] = true
   if self.Config.AchievementTrackerMode == ModeRecentlyChangedId then
    self:BuildUpAchievementList()
    updateSingelItem = false
   end
  end
  -- Only update data in one single item.
  if updateSingelItem == true then
   local item = self.AchievmentItemList[achievementId]
   if item ~= nil then
    -- Make sure the item is visible.
    item:GetParent():EnsureChildVisible(item)
    self:UpdateAchievementItem(item)
   end
  end
 end
end

------------------------------------------------------------------------------------------------------
-- AchievementTracker OnAchievementUpdateTimer
-- Validate if a achievement need additional update.
------------------------------------------------------------------------------------------------------
function AchievementTracker:OnAchievementUpdateTimer()
 for position, item in pairs(self.AchievmentItemList) do
  if item ~= nil then
   local data = item:GetData()
   local progressValue = data:GetNumCompleted()
   local progressBar = item:FindChild("AchievementProgressBar")
   local progressBarValue = progressBar:GetData()
   if progressValue ~= progressBarValue then
    -- Post info for delayed achievements in debug mode.
    if ShowDebugInfos == true then
     Print(
      "Found for Update - \nName: " .. data:GetName() ..
      "\nId: " .. data:GetId() ..
      " Changed value: " .. progressBarValue ..
      " to " .. progressValue)
    end
    -- This item have a missing change, so do update for them.
    item:GetParent():EnsureChildVisible(item)
    self:UpdateAchievementItem(item)
   end
  end
 end
end

------------------------------------------------------------------------------------------------------
-- AchievementTracker OnConfigure
-- Open the options form if configuration is called.
------------------------------------------------------------------------------------------------------
function AchievementTracker:OnConfigure()
 self:OnButtonSignalOptions()
end

------------------------------------------------------------------------------------------------------
-- AchievementTracker OnDocLoaded
-- The xmlDoc will not set to nil, because the achievement tracker generates dynamic list items.
------------------------------------------------------------------------------------------------------
function AchievementTracker:OnDocLoaded()
 if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
 self.MainForm = Apollo.LoadForm(self.xmlDoc, "AchievementTrackerForm", nil, self)
 self.OptionsForm = Apollo.LoadForm(self.xmlDoc, "AchievementTrackerOptionsForm", nil, self)
  if self.MainForm == nil or
     self.OptionsForm == nil then
   Apollo.AddAddonErrorText(self, "Could not load forms for some reason.")
  else
   -- For preparing data the forms are generally and immediately hidden.
   self.MainForm:Show(false, true)
   self.OptionsForm:Show(false, true)
   if self.Config ~= nil then
    -- Restore saved window positions.
    if self.Config.PositionMain ~= nil then
     self.MainForm:SetAnchorOffsets(
      self.Config.PositionMain.Left,
      self.Config.PositionMain.Top,
      self.Config.PositionMain.Right,
      self.Config.PositionMain.Bottom)
    end
    if self.Config.PositionOptions ~= nil then
     self.OptionsForm:SetAnchorOffsets(
      self.Config.PositionOptions.Left,
      self.Config.PositionOptions.Top,
      self.Config.PositionOptions.Right,
      self.Config.PositionOptions.Bottom)
    end
   end

   -- Register slash commands to open the main window.
   Apollo.RegisterSlashCommand("achvtrack", "OnAchievementTrackerCall", self)
   Apollo.RegisterEventHandler("AchievementTrackerMenuClicked", "OnAchievementTrackerCall", self)
   -- Register to add entries to window management.
   Apollo.RegisterEventHandler("WindowManagementReady", "OnWindowManagementReady", self)

   -- The timers AchievementUpdateTimer and AchievementBuildUpTimer all hacks to solve problems. Possibly if they fixed, then they can remove.
   -- Prepare timer to update achievement items when they got no update via the event options.
   Apollo.RegisterTimerHandler("AchievementUpdateTimer", "OnAchievementUpdateTimer", self)
   Apollo.CreateTimer("AchievementUpdateTimer", 5, true)
   Apollo.StopTimer("AchievementUpdateTimer")
   -- Workaround to catch all data, because after first ui load not all data are avaiable. Trigger one reload short after loading this addon.
   Apollo.RegisterTimerHandler("AchievementBuildUpTimer", "BuildUpAchievementList", self)
   Apollo.CreateTimer("AchievementBuildUpTimer", 1, false)
   Apollo.StartTimer("AchievementBuildUpTimer")

   -- Set ModeItemArea as dropdownlist for button ModeComboBox
   self.OptionsForm:FindChild("ModeComboBox"):AttachWindow(self.OptionsForm:FindChild("ModeItemArea"))

   -- Open the main form if it was visible before.
   if self.Config ~= nil and
      self.Config.IsVisible == true then
    self:MainShow()
   end
   -- If Debug is on, print a warning.
   if ShowDebugInfos == true then
    Print("Warning: Debug mode active. Plase deactivate for release.")
   end
  end -- self.MainForm == nil / self.OptionsForm == nil
 end -- self.xmlDoc ~= nil
end

------------------------------------------------------------------------------------------------------
-- AchievementTracker OnInterfaceMenuListHasLoaded
-- Adds a menu entry for this addon.
------------------------------------------------------------------------------------------------------
function AchievementTracker:OnInterfaceMenuListHasLoaded()
 Event_FireGenericEvent(
  "InterfaceMenuList_NewAddOn",
  "Achievement Tracker",
  {
   "AchievementTrackerMenuClicked",
   "",
   "Icon_Windows32_UI_CRB_InterfaceMenu_Achievements"
  })
end

------------------------------------------------------------------------------------------------------
-- AchievementTracker OnLoad
-- Load the form xml file.
------------------------------------------------------------------------------------------------------
function AchievementTracker:OnLoad()
 self.xmlDoc = XmlDoc.CreateFromFile("AchievementTracker.xml")
 self.xmlDoc:RegisterCallback("OnDocLoaded", self)
 -- Register for the InterfaceMenuListHasLoaded event to add a own menu item.
 Apollo.RegisterEventHandler("InterfaceMenuListHasLoaded", "OnInterfaceMenuListHasLoaded", self)
end

------------------------------------------------------------------------------------------------------
-- AchievementTracker OnRestore
-- Load the saved settings for the addon.
------------------------------------------------------------------------------------------------------
function AchievementTracker:OnRestore(type, savedData)
 if type == GameLib.CodeEnumAddonSaveLevel.Character then
  if savedData ~= nil and
     savedData.Version ~= nil then
   if savedData.Version == 1 then
    -- Migration path from version 1 to version 2. Use the defaults and create realm list.
    savedData.NegativeCategoryFilter = self.Config.NegativeCategoryFilter
    savedData.Version = 2
   end
   if savedData.Version >= ConfigMinVersion then
    -- Saved configs only used if they compatible with the minimum version.
    self.Config = savedData
   end
  end
 end
end

------------------------------------------------------------------------------------------------------
-- AchievementTracker OnSave
-- Save the settings for the addon per Character.
------------------------------------------------------------------------------------------------------
function AchievementTracker:OnSave(eType)
 local result = nil
 if eType == GameLib.CodeEnumAddonSaveLevel.Character then
  result = self.Config
 end
 return result
end

------------------------------------------------------------------------------------------------------
-- AchievementTracker OnWindowManagementReady
-- Add forms to addon window management.
------------------------------------------------------------------------------------------------------
function AchievementTracker:OnWindowManagementReady()
 Event_FireGenericEvent("WindowManagementAdd", {wnd = self.MainForm, strName = "Achievement Tracker"})
 Event_FireGenericEvent("WindowManagementAdd", {wnd = self.OptionsForm, strName = "Achievement Tracker Options"})
end

--****************************************************************************************************
-- Eventfunctions for AchievementTrackerForm 
--****************************************************************************************************
------------------------------------------------------------------------------------------------------
-- AchievementTrackerForm OnButtonSignalClose
-- Close function for the AchievementTrackerForm.
------------------------------------------------------------------------------------------------------
function AchievementTracker:OnButtonSignalClose( wndHandler, wndControl, eMouseButton )
 self:MainHide()
end

------------------------------------------------------------------------------------------------------
-- AchievementTrackerOptionsForm OnButtonSignalModeComboBox
-- Open the drop down for combobox.
------------------------------------------------------------------------------------------------------
function AchievementTracker:OnButtonSignalModeComboBox( wndHandler, wndControl, eMouseButton )
 self.OptionsForm:FindChild("ModeComboBox"):SetCheck(true)
end

------------------------------------------------------------------------------------------------------
-- AchievementTrackerForm OnButtonSignalOptions
-- Open the configuration form.
------------------------------------------------------------------------------------------------------
function AchievementTracker:OnButtonSignalOptions( wndHandler, wndControl, eMouseButton )
 if self.OptionsForm:IsShown() == false then
  self:BuildUpOptionsLists()
 end
 self.OptionsForm:Show(not self.OptionsForm:IsShown(), false)
end

------------------------------------------------------------------------------------------------------
-- AchievementTrackerForm OnWindowMove
-- Saves the position after moving window in config data.
------------------------------------------------------------------------------------------------------
function AchievementTracker:OnWindowMoveMain( wndHandler, wndControl, nOldLeft, nOldTop, nOldRight, nOldBottom )
 self.Config.PositionMain = self:GetPosition(self.MainForm)
end

--****************************************************************************************************
-- Eventfunctions for AchievementTrackerOptionsForm
--****************************************************************************************************
------------------------------------------------------------------------------------------------------
-- AchievementTrackerOptionsForm OnButtonCheckNegativeCategoryFilter
------------------------------------------------------------------------------------------------------
function AchievementTracker:OnButtonCheckNegativeCategoryFilter( wndHandler, wndControl, eMouseButton )
 self.Config.NegativeCategoryFilter = true
 if self.Config ~= nil and
    self.Config.AchievementTrackerMode == ModeFilterId then
  self:BuildUpAchievementList()
 end
end

------------------------------------------------------------------------------------------------------
-- AchievementTrackerOptionsForm OnButtonCheckShowCompletedAchievements
------------------------------------------------------------------------------------------------------
function AchievementTracker:OnButtonCheckShowCompletedAchievements( wndHandler, wndControl, eMouseButton )
 self.Config.ShowCompletedAchievements = true
 if self.Config ~= nil and
    self.Config.AchievementTrackerMode == ModeFilterId then
  self:BuildUpAchievementList()
 end
end

------------------------------------------------------------------------------------------------------
-- AchievementTrackerOptionsForm OnButtonCheckShowGuildAchievements
------------------------------------------------------------------------------------------------------
function AchievementTracker:OnButtonCheckShowGuildAchievements( wndHandler, wndControl, eMouseButton )
 self.Config.ShowGuildAchievements = true
 if self.Config ~= nil and
    self.Config.AchievementTrackerMode == ModeFilterId then
  self:BuildUpAchievementList()
 end
end

------------------------------------------------------------------------------------------------------
-- AchievementTrackerOptionsForm OnButtonUncheckNegativeCategoryFilter
------------------------------------------------------------------------------------------------------
function AchievementTracker:OnButtonUncheckNegativeCategoryFilter( wndHandler, wndControl, eMouseButton )
 self.Config.NegativeCategoryFilter = false
 if self.Config ~= nil and
    self.Config.AchievementTrackerMode == ModeFilterId then
  self:BuildUpAchievementList()
 end
end

------------------------------------------------------------------------------------------------------
-- AchievementTrackerOptionsForm OnButtonUncheckShowCompletedAchievements
------------------------------------------------------------------------------------------------------
function AchievementTracker:OnButtonUncheckShowCompletedAchievements( wndHandler, wndControl, eMouseButton )
 self.Config.ShowCompletedAchievements = false
 if self.Config ~= nil and
    self.Config.AchievementTrackerMode == ModeFilterId then
  self:BuildUpAchievementList()
 end
end

------------------------------------------------------------------------------------------------------
-- AchievementTrackerOptionsForm OnButtonUncheckShowGuildAchievements
------------------------------------------------------------------------------------------------------
function AchievementTracker:OnButtonUncheckShowGuildAchievements( wndHandler, wndControl, eMouseButton )
 self.Config.ShowGuildAchievements = false
 if self.Config ~= nil and
    self.Config.AchievementTrackerMode == ModeFilterId then
  self:BuildUpAchievementList()
 end
end

------------------------------------------------------------------------------------------------------
-- AchievementTrackerOptionsForm OnButtonSignalAddToCustomList
-- Add the current item to the custom list.
------------------------------------------------------------------------------------------------------
function AchievementTracker:OnButtonSignalAddToCustomList( wndHandler, wndControl, eMouseButton )
 if wndHandler ~= nil then
  data = wndHandler:GetParent():GetData()
  self.Config.CustomList[data:GetId()] = true
 end
end

------------------------------------------------------------------------------------------------------
-- AchievementTrackerOptionsForm OnButtonSignalRemoveFromCustomList
-- Removes the current item from the custom list.
------------------------------------------------------------------------------------------------------
function AchievementTracker:OnButtonSignalRemoveFromCustomList( wndHandler, wndControl, eMouseButton )
 if wndHandler ~= nil then
  local parent = wndHandler:GetParent()
  local data = parent:GetData()
  self.Config.CustomList[data:GetId()] = nil
  parent:Destroy()
  self.MainForm:FindChild("ItemContainer"):ArrangeChildrenVert()
 end
end

------------------------------------------------------------------------------------------------------
-- AchievementTrackerOptionsForm OnButtonSignalClearCustomList
-- Removes all items in the custom achievement list.
------------------------------------------------------------------------------------------------------
function AchievementTracker:OnButtonSignalClearCustomList( wndHandler, wndControl, eMouseButton )
 self.Config.CustomList = {}
 if self.Config ~= nil and
    self.Config.AchievementTrackerMode == ModeCustomListId then
  self:BuildUpAchievementList()
 end
end

------------------------------------------------------------------------------------------------------
-- AchievementTrackerOptionsForm OnButtonSignalClearFilterList
-- Removes all items in area and category filter list.
------------------------------------------------------------------------------------------------------
function AchievementTracker:OnButtonSignalClearFilterList( wndHandler, wndControl, eMouseButton )
 self.Config.FilterCategory = {}
 self.Config.FilterZone = {}
 self:BuildUpOptionsLists()
 if self.Config ~= nil and
    self.Config.AchievementTrackerMode == ModeFilterId then
  self:BuildUpAchievementList()
 end
end

------------------------------------------------------------------------------------------------------
-- AchievementTrackerOptionsForm OnButtonSignalClearRecentlyChangedList
-- Removes all items in the recently changed achievement list.
------------------------------------------------------------------------------------------------------
function AchievementTracker:OnButtonSignalClearRecentlyChangedList( wndHandler, wndControl, eMouseButton )
 self.Config.RecentlyChangedList = {}
 if self.Config ~= nil and
    self.Config.AchievementTrackerMode == ModeRecentlyChangedId then
  self:BuildUpAchievementList()
 end
end

------------------------------------------------------------------------------------------------------
-- AchievementTrackerOptionsForm OnButtonSignalCloseOptions
-- Close function for the AchievementTrackerOptionsForm.
------------------------------------------------------------------------------------------------------
function AchievementTracker:OnButtonSignalCloseOptions( wndHandler, wndControl, eMouseButton )
 self:OnButtonSignalOptions(wndHandler, wndControl, eMouseButton)
end

------------------------------------------------------------------------------------------------------
-- AchievementTrackerOptionsForm OnTreeDoubleClickCategory
-- Add or removes the node from the category filter list.
------------------------------------------------------------------------------------------------------
function AchievementTracker:OnTreeDoubleClickCategory( wndHandler, wndControl, hNode )
 local tree = self.OptionsForm:FindChild("CategoryTreeControl")
 local data = tree:GetNodeData(hNode)
 if data ~= nil then
  if data.IsInFilter == true then
   self.Config.FilterCategory[data.Id] = nil
  else
   self.Config.FilterCategory[data.Id] = true
  end
  data.IsInFilter = not data.IsInFilter
  tree:SetNodeImage(hNode, self:GetTreeImage(data.IsInFilter))
  if self.Config ~= nil and
     self.Config.AchievementTrackerMode == ModeFilterId then
   self:BuildUpAchievementList()
  end
 end
end

------------------------------------------------------------------------------------------------------
-- AchievementTrackerOptionsForm OnTreeDoubleClickZone
-- Add or removes the node from the zone filter list.
------------------------------------------------------------------------------------------------------
function AchievementTracker:OnTreeDoubleClickZone( wndHandler, wndControl, hNode )
 local tree = self.OptionsForm:FindChild("ZoneTreeControl")
 local data = tree:GetNodeData(hNode)
 if data ~= nil then
  for position, item in pairs(data.Id) do
   if data.IsInFilter == true then
    self.Config.FilterZone[item] = nil
   else
    self.Config.FilterZone[item] = true
   end
  end
  data.IsInFilter = not data.IsInFilter
  tree:SetNodeImage(hNode, self:GetTreeImage(data.IsInFilter))
  if self.Config ~= nil and
     self.Config.AchievementTrackerMode == ModeFilterId then
   self:BuildUpAchievementList()
  end
 end
end

------------------------------------------------------------------------------------------------------
-- AchievementTrackerOptionsForm OnWindowMoveOptions
-- Saves the position after moving window in config data.
------------------------------------------------------------------------------------------------------
function AchievementTracker:OnWindowMoveOptions( wndHandler, wndControl, nOldLeft, nOldTop, nOldRight, nOldBottom )
 self.Config.PositionOptions = self:GetPosition(self.OptionsForm)
 -- Task for size changed handled in the move event as workaround to prevent consuming size changed
 -- events to prevent performance issues because the triggered to often.
 self.OptionsForm:FindChild("CategoryTreeControl"):SetColumnWidth(1, (self.OptionsForm:GetWidth() - 70) / 2)
 self.OptionsForm:FindChild("ZoneTreeControl"):SetColumnWidth(1, (self.OptionsForm:GetWidth() - 70) / 2)
end

--****************************************************************************************************
-- Eventfunctions for ModeListItem
--****************************************************************************************************
------------------------------------------------------------------------------------------------------
-- ModeListItem OnButtonSignalModeListItem
-- Change the mode of the achievement tracker.
------------------------------------------------------------------------------------------------------
function AchievementTracker:OnButtonSignalModeListItem( wndHandler, wndControl, eMouseButton )
 if wndHandler ~= nil then
  local mode = wndHandler:GetData()
  self.Config.AchievementTrackerMode = mode.Id
  self:UpdateModeComboBox(mode)
  -- Rebuild achivmentlist.
  self:BuildUpAchievementList()
  -- Close the combobox itemlist.
  self.OptionsForm:FindChild("ModeComboBox"):SetCheck(false)
 end
end

--****************************************************************************************************
-- AchievementTracker Helper Functions
--****************************************************************************************************
------------------------------------------------------------------------------------------------------
-- AchievementTracker AddListToTree
-- Adds the items and childs to the tree control.
------------------------------------------------------------------------------------------------------
function AchievementTracker:AddListToTree(parent, tree, treeData)
 if parent ~= nil and
    tree ~= nil and
    treeData ~= nil then
  -- Iterate the category items with Child, Id, IsInFilter and Name.
  for position, item in pairs(treeData) do
   local node = tree:AddNode(parent, item.Name, "", item)
   tree:SetNodeImage(node, self:GetTreeImage(item.IsInFilter))
   self:AddListToTree(node, tree, item.ChildList)
   if item.HideChilds == true then
    tree:CollapseNode(node)
   end
  end
 end
end

------------------------------------------------------------------------------------------------------
-- AchievementTracker BuildUpOptionsLists
-- This window build up the options list in the options form.
------------------------------------------------------------------------------------------------------
function AchievementTracker:BuildUpOptionsLists()
 if self.OptionsForm ~= nil then
  -- Restore options from config.
  self.OptionsForm:FindChild("ShowCompletedAchievements"):SetCheck(self.Config.ShowCompletedAchievements)
  self.OptionsForm:FindChild("ShowGuildAchievements"):SetCheck(self.Config.ShowGuildAchievements)
  -- Build up the items for mode selection.
  local parent = self.OptionsForm:FindChild("ModeItemArea")
  parent:DestroyChildren()
  for position, item in pairs(AchievementTrackerMode) do
   local modeItem = Apollo.LoadForm(self.xmlDoc, "ModeListItem", parent, self)
   modeItem:SetTooltip(item.Tooltip)
   local modeButton = modeItem:FindChild("ModeButton")
   modeButton:SetData(item)
   modeButton:SetText(item.Name)
   modeButton:SetTooltip(item.Tooltip)
   if self.Config.AchievementTrackerMode == item.Id then
    -- The current mode will selected as value for the drop down.
    self:UpdateModeComboBox(item)
   end
  end
  parent:ArrangeChildrenVert()
  -- Build up the items for category filter.
  local tree = self.OptionsForm:FindChild("CategoryTreeControl")
  tree:DeleteAll()
  local treeData = self:GetAchievementCategoryTree()
  self:AddListToTree(0, tree, treeData)
  -- Build up the items for zone filter.
  tree = self.OptionsForm:FindChild("ZoneTreeControl")
  tree:DeleteAll()
  local treeData = self:GetAchievementZoneList()
  self:AddListToTree(0, tree, treeData)
 end
end

------------------------------------------------------------------------------------------------------
-- AchievementTracker BuildUpAchievementList
-- This function shows all achievments for the selected mode and filter.
------------------------------------------------------------------------------------------------------
function AchievementTracker:BuildUpAchievementList()
 local achievementList = self:GetAchievementList()
 local parent = self.MainForm:FindChild("ItemContainer")
 self:UpdateTitel()
 self.AchievmentItemList = {}
 parent:DestroyChildren()
 -- Call this to reset the scroll bar.
 parent:ArrangeChildrenVert()
 for position, item in pairs(achievementList) do
  local achievementItem = self:GetAchievementItem(item, parent)
  -- Add items to the internal list for direct updates.
  self.AchievmentItemList[item:GetId()] = achievementItem
 end
 parent:ArrangeChildrenVert()
end

------------------------------------------------------------------------------------------------------
-- AchievementTracker GetAchievementCategoryIcon
-- Returns the icon for the given category.
-- Avaiable Icons (* = used)
-- *IconSprites:Icon_Achievement_Achievement_Adventures - Id: 78
-- *IconSprites:Icon_Achievement_Achievement_Challenges - Id: 161
-- *IconSprites:Icon_Achievement_Achievement_Combat - Id: 4
-- IconSprites:Icon_Achievement_Achievement_Datacube
-- *IconSprites:Icon_Achievement_Achievement_Dungeon - Id: 65
-- IconSprites:Icon_Achievement_Achievement_Equipment
-- *IconSprites:Icon_Achievement_Achievement_Exploration - Id: 1
-- *IconSprites:Icon_Achievement_Achievement_GalacticArchives - Id: 175
-- *IconSprites:Icon_Achievement_Achievement_GenericAchievement - Id: 12
-- IconSprites:Icon_Achievement_Achievement_Guild
-- IconSprites:Icon_Achievement_Achievement_HousingExterior
-- IconSprites:Icon_Achievement_Achievement_HousingInterior
-- IconSprites:Icon_Achievement_Achievement_Journals
-- IconSprites:Icon_Achievement_Achievement_Legacy
-- IconSprites:Icon_Achievement_Achievement_MetaAchievement
-- *IconSprites:Icon_Achievement_Achievement_Path_Explorer - Id: 9
-- *IconSprites:Icon_Achievement_Achievement_Path_Scientist - Id: 9
-- *IconSprites:Icon_Achievement_Achievement_Path_Settler - Id: 9
-- *IconSprites:Icon_Achievement_Achievement_Path_Soldier - Id: 9
-- IconSprites:Icon_Achievement_Achievement_PermGuildAmenities
-- IconSprites:Icon_Achievement_Achievement_PermVendorUnlocks
-- *IconSprites:Icon_Achievement_Achievement_PvP - Id: 7
-- *IconSprites:Icon_Achievement_Achievement_Quest - Id: 3
-- *IconSprites:Icon_Achievement_Achievement_Raid - Id: 164
-- *IconSprites:Icon_Achievement_Achievement_Reputation - Id: 11
-- IconSprites:Icon_Achievement_Achievement_ServerWide_Explorer
-- *IconSprites:Icon_Achievement_Achievement_ServerWide_General - Id: 254
-- IconSprites:Icon_Achievement_Achievement_ServerWide_Scientist
-- IconSprites:Icon_Achievement_Achievement_ServerWide_Settler
-- IconSprites:Icon_Achievement_Achievement_ServerWide_Soldier
-- *IconSprites:Icon_Achievement_Achievement_Shiphand - Id: 162
-- *IconSprites:Icon_Achievement_Achievement_Social - Id: 281
-- IconSprites:Icon_Achievement_Achievement_SocialCircles
-- IconSprites:Icon_Achievement_Achievement_Tales
-- IconSprites:Icon_Achievement_Achievement_TempGuildAmenities
-- IconSprites:Icon_Achievement_Achievement_Tradeskill_Architect
-- IconSprites:Icon_Achievement_Achievement_Tradeskill_Armorer
-- IconSprites:Icon_Achievement_Achievement_Tradeskill_Cooking
-- IconSprites:Icon_Achievement_Achievement_Tradeskill_Farmer
-- IconSprites:Icon_Achievement_Achievement_Tradeskill_Miner
-- IconSprites:Icon_Achievement_Achievement_Tradeskill_Outfitter
-- IconSprites:Icon_Achievement_Achievement_Tradeskill_RelicHunter
-- IconSprites:Icon_Achievement_Achievement_Tradeskill_Survivalist
-- IconSprites:Icon_Achievement_Achievement_Tradeskill_Tailor
-- IconSprites:Icon_Achievement_Achievement_Tradeskill_Technologist
-- IconSprites:Icon_Achievement_Achievement_Tradeskill_WeaponCrafting
-- IconSprites:Icon_Achievement_Achievement_WeeklyGuildBuff
-- *IconSprites:Icon_Achievement_Achievement_WorldEvent - Id: 163
-- IconSprites:Icon_Achievement_Achievement_Zone
------------------------------------------------------------------------------------------------------
function AchievementTracker:GetAchievementCategoryIcon(categoryId)
 local result = ""
 if categoryId == 9 then -- Pfade
  local playerPath = PlayerPathLib.GetPlayerPathType()
  -- The carbine addon used only IconSprites:Icon_Achievement_Achievement_Path_Explorer.
  if playerPath == 0 then
   result = "IconSprites:Icon_Achievement_Achievement_Path_Soldier"
  elseif playerPath == 1 then
   result = "IconSprites:Icon_Achievement_Achievement_Path_Settler"
  elseif playerPath == 2 then
   result = "IconSprites:Icon_Achievement_Achievement_Path_Scientist"
  elseif playerPath == 3 then
   result = "IconSprites:Icon_Achievement_Achievement_Path_Explorer"
  end
 else
  result = AchievementIconCategoryMapping[categoryId]
 end
 return result
end

------------------------------------------------------------------------------------------------------
-- AchievementTracker GetAchievementCategoryTree
-- Returns the category tree converted for recursive using.
------------------------------------------------------------------------------------------------------
function AchievementTracker:GetAchievementCategoryTree()
 local result = {}
 local treeData = AchievementsLib.GetAchievementCategoryTree()
 table.sort(treeData, function(a,b) return a.strCategoryName < b.strCategoryName end)
 -- Iterate category list with nCategoryId, strCategoryName, tGroups.
 for position, item in pairs(treeData) do
  local data =
  {
   ChildList = {},
   HideChilds = false,
   Id = item.nCategoryId,
   IsInFilter = self.Config.FilterCategory[item.nCategoryId] ~= nil,
   Name = item.strCategoryName,
  }
  table.insert(result, data)
  -- Iterate groups in the category with nGroupId, strGroupName, tSubGroups
  if table.getn(item.tGroups) > 0 then
   table.sort(item.tGroups, function(a,b) return a.strGroupName < b.strGroupName end)
   for positionGroup, itemGroup in pairs(item.tGroups) do
    local dataGroup =
    {
     ChildList = {},
     HideChilds = true,
     Id = itemGroup.nGroupId,
     IsInFilter = self.Config.FilterCategory[itemGroup.nGroupId] ~= nil,
     Name = itemGroup.strGroupName,
    }
    table.insert(data.ChildList, dataGroup)
    -- Iterate subgroups in groups with nSubGroupId, strSubGroupName
    if table.getn(itemGroup.tSubGroups) > 0 then
     table.sort(itemGroup.tSubGroups, function(a,b) return a.strSubGroupName < b.strSubGroupName end)
     for positionSubGroup, itemSubGroup in pairs(itemGroup.tSubGroups) do
      local dataSubGroup =
      {
       ChildList = {},
       HideChilds = true,
       Id = itemSubGroup.nSubGroupId,
       IsInFilter = self.Config.FilterCategory[itemSubGroup.nSubGroupId] ~= nil,
       Name = itemSubGroup.strSubGroupName,
      }
      table.insert(dataGroup.ChildList, dataSubGroup)
     end -- for tSubGroups
    end -- if
   end -- for tGroups
  end -- if
 end -- for treeData
 return result
end

------------------------------------------------------------------------------------------------------
-- AchievementTracker GetAchievementItem
-- Returns a new achievement item.
------------------------------------------------------------------------------------------------------
function AchievementTracker:GetAchievementItem(data, parent)
 local result = Apollo.LoadForm(self.xmlDoc, "AchievementItem", parent, self)
 result:SetData(data)
 self:UpdateAchievementItem(result)
 return result
end

------------------------------------------------------------------------------------------------------
-- AchievementTracker GetAchievementList
-- Returns the list of achievements for the current mode and filter.
-- The list is sorted by name, but the parent and child achievements are bound together to stay
-- behind.
------------------------------------------------------------------------------------------------------
function AchievementTracker:GetAchievementList()
 local result = {}
 -- For preparation the sorting all parent achievements bound in one list.
 local achievementParentList = {}
 -- Iterate the personal achievements.
 for position, item in pairs(AchievementsLib.GetAchievements(false)) do
  if item:GetParentTier() == nil then
   table.insert(achievementParentList, item)
  end
 end
 -- Now iterate the guild achievements.
 for position, item in pairs(AchievementsLib.GetAchievements(true)) do
  if item:GetParentTier() == nil then
   table.insert(achievementParentList, item)
  end
 end
 table.sort(achievementParentList, function(a,b) return a:GetName() < b:GetName() end)
 -- Now iterate the parent list and the child items to filter them.
 for position, item in pairs(achievementParentList) do
  local itemToAdd = item
  while itemToAdd ~= nil do
   local itemId = itemToAdd:GetId()
   local addItem = false
   if itemToAdd:GetName() ~= "" then
    if self.Config.AchievementTrackerMode == ModeAllId then
     -- Return all achievments.
     addItem = true
    elseif self.Config.AchievementTrackerMode == ModeCustomListId then
     -- Return the achievements that id is in custom list.
     addItem = self.Config.CustomList[itemId] ~= nil
    elseif self.Config.AchievementTrackerMode == ModeFilterId then
     -- Return only this achievements that match te current filter criterias.
     local isCompleted = itemToAdd:IsComplete()
     local isGuild = itemToAdd:IsGuildAchievement()
     local isCategoryFilter = self:GetTablelength(self.Config.FilterCategory) > 0
     local isZoneFilter = self:GetTablelength(self.Config.FilterZone) > 0
     local filterCompleteOk = false
     local filterGuildOk = false
     local filterCategoryOk = false
     local filterZoneOk = false
     if isCompleted == false or
       (isCompleted == true and self.Config.ShowCompletedAchievements == true) then
      filterCompleteOk = true
     end
     if isGuild == false or
       (isCompleted == true and self.Config.ShowGuildAchievements == true) then
      filterGuildOk = true
     end
     if isCategoryFilter == false or
        (isCategoryFilter == true and
         self.Config.NegativeCategoryFilter == false and
         self.Config.FilterCategory[itemToAdd:GetCategoryId()] ~= nil) or
        (isCategoryFilter == true and
         self.Config.NegativeCategoryFilter == true and
         self.Config.FilterCategory[itemToAdd:GetCategoryId()] == nil) then
      filterCategoryOk = true
     end
     if isZoneFilter == false or
       (isZoneFilter == true and self.Config.FilterZone[itemToAdd:GetWorldZoneId()] ~= nil) then
      filterZoneOk = true
     end
     addItem = filterCompleteOk and filterGuildOk and filterCategoryOk and filterZoneOk
    elseif self.Config.AchievementTrackerMode == ModeRecentlyChangedId then
     -- Return the achievements that id is in recently changed list.
     addItem = self.Config.RecentlyChangedList[itemId] ~= nil
    end
    if addItem == true then
     table.insert(result, itemToAdd)
    end
   end -- if GetName
   itemToAdd = itemToAdd:GetChildTier()
  end -- while itemToAdd ~= nil do
 end -- for position, item in pairs(achievementParentList) do
 return result
end

------------------------------------------------------------------------------------------------------
-- AchievementTracker GetAchievementNamePrefix
-- Returns a prefix depending on the achivement. Default is "".
-- Prefix for guild achievements is "(G) ".
------------------------------------------------------------------------------------------------------
function AchievementTracker:GetAchievementNamePrefix(achievement)
 local result = ""
  if achievement ~= nil and
     achievement:IsGuildAchievement() == true then
   result = NamePrefixGuild
  end
 return result
end

------------------------------------------------------------------------------------------------------
-- AchievementTracker GetAchievementItem
-- Returns a tooltip for a achievement item.
------------------------------------------------------------------------------------------------------
function AchievementTracker:GetAchievementTooltip(data)
 local result = "No data avaiable."
 if data ~= nil then
  local checkListInfo = ""
  if data:IsChecklist() == true then
   -- Combine a string with the name of the items.
   checkListInfo = checkListInfo .. "\n "
   for position, item in pairs(data:GetChecklistItems()) do
    if item.bIsComplete == true then
     checkListInfo = checkListInfo .. "\n+ " .. item.strChecklistEntry
    else
     checkListInfo = checkListInfo .. "\n- " .. item.strChecklistEntry
    end
   end
  end
  if ShowDebugInfos == true then
   result = "Titel: " .. data:GetName()
    .. "\nDescription\n" .. data:GetDescription() .. "\n \n"
    .. "\nCategory: " .. data:GetCategory()
    .. "\nCategoryFullName: " .. data:GetCategoryFullName()
    .. "\nCategoryId: " .. data:GetCategoryId()
    .. "\nGetChecklistItems Exist: " .. tostring(data:GetChecklistItems() ~= nil)
    .. "\nChecklistItems" .. checkListInfo .. "\n \n"
    .. "\nChildTier Exist: " .. tostring(data:GetChildTier() ~= nil)
    .. "\nCurrencySystem: " .. tostring(data:GetCurrencySystem())     
    .. "\nDateCompleted: " .. data:GetDateCompleted()
    .. "\nGroup: " .. data:GetGroup()
    .. "\nId: " .. data:GetId()
    .. "\nNumCompleted: " .. data:GetNumCompleted()
    .. "\nNumNeeded: " .. data:GetNumNeeded()
    .. "\nParentTier Exist: " .. tostring(data:GetParentTier() ~= nil)
    .. "\nPoints: " .. data:GetPoints()
    .. "\nProgressText\n" .. data:GetProgressText() .. "\n \n"
    .. "\nReward Entries: " .. table.getn(data:GetRewards())
    .. "\nSubGroup: " .. data:GetSubGroup()
    .. "\nSubGroupId: " .. data:GetSubGroupId()
    .. "\nTradeskillLayout Exist: " .. tostring(data:GetTradeskillLayout() ~= nil)
    .. "\nTradeskillRewards Exist: " .. tostring(data:GetTradeskillRewards() ~= nil)
    .. "\nTradeskillSchematicsRequired Entries: " .. table.getn(data:GetTradeskillSchematicsRequired())
    .. "\nWorldZoneId: " .. data:GetWorldZoneId()
    .. "\nIsChecklist: " .. tostring(data:IsChecklist())
    .. "\nIsComplete: " .. tostring(data:IsComplete())
    .. "\nIsCurrencyShown: " .. tostring(data:IsCurrencyShown())
    .. "\nIsGuildAchievement: " .. tostring(data:IsGuildAchievement())
  else
   result = ""
   if data:IsComplete() == true then
    result = data:GetName() ..
     "\n \nCompleted: " .. data:GetDateCompleted() ..
     "\n \nDescription\n" .. data:GetDescription() ..
     checkListInfo
   else
    result = data:GetName() ..
     "\n \nTask\n" .. data:GetProgressText() ..
     checkListInfo
   end
  end
 end
 return result
end

------------------------------------------------------------------------------------------------------
-- AchievementTracker GetAchievementZoneList
-- Returns the zonelist for achievements with added names to empty items and goup by names.
-- Namechanges: 12 = *World, 140 = *Arenas, empty name = ?Not named
------------------------------------------------------------------------------------------------------
function AchievementTracker:GetAchievementZoneList()
 local result = {}
 local zoneList = AchievementsLib.GetAchievementZones()
 local zoneWithAchievementList = {}
 -- Prepare name changes before sort.
 for position, item in pairs(zoneList) do
  -- Validate that for a region achievements exist or remove it.
  if self:ZoneHasAchievments(item.nId) == true then
   if item.strName == "" then
    -- Name some Areas that have empty names.
    if item.nId == 12 then
     item.strName = "*World"
    elseif item.nId == 140 then
     item.strName = "*Arenas"
    else
     item.strName = "?Not named"
    end
   end
   table.insert(zoneWithAchievementList, item)
  end
 end
 table.sort(zoneWithAchievementList, function(a,b) return a.strName < b.strName end)
 -- Because the sorted names will be lost, if the items add with names, so a helper list will used.
 local dataList = {}
 -- Iterate zone list with nId and strName.
 for position, item in pairs(zoneWithAchievementList) do
  local data = dataList[item.strName]
  if data ~= nil then
   -- This zone name already exist in the result, so only add the id to the list.
   table.insert(data.Id, item.nId)
  else
   -- This zone is not in the result before, so it will created and added.
   data =
   {
    ChildList = {},
    HideChilds = false,
    Id =
    {
     item.nId,
    },
    IsInFilter = self.Config.FilterZone[item.nId] ~= nil,
    Name = item.strName,
   }
   dataList[item.strName] = data
   table.insert(result, data)
  end
 end
 return result
end

------------------------------------------------------------------------------------------------------
-- AchievementTracker GetPosition
-- Returns the position of the form as table.
------------------------------------------------------------------------------------------------------
function AchievementTracker:GetPosition(form)
 local result = nil
 if form ~= nil then
  local leftAnchor, topAnchor, rightAnchor, bottomAnchor = form:GetAnchorOffsets()
  result =
  {
   Bottom = bottomAnchor,
   Left = leftAnchor,
   Right = rightAnchor,
   Top = topAnchor,
  }
 end
 return result
end

------------------------------------------------------------------------------------------------------
-- AchievementTracker GetTablelength
-- Function to count the items in a table, because table.getn() is not working for named items.
------------------------------------------------------------------------------------------------------
function AchievementTracker:GetTablelength(table)
 local result = 0
 for _ in pairs(table) do
  result = result + 1
 end
 return result
end

------------------------------------------------------------------------------------------------------
-- AchievementTracker GetTreeImage
-- Returns the image for the tree item, depending if the item is in a list or not.
-- True = BK3:btnHolo_CheckPressed
-- False = BK3:btnHolo_CheckDisabled
------------------------------------------------------------------------------------------------------
function AchievementTracker:GetTreeImage(state)
 local result = "BK3:btnHolo_CheckDisabled"
 if state == true then
  result = "BK3:btnHolo_CheckPressed"
 end
 return result
end

------------------------------------------------------------------------------------------------------
-- AchievementTracker MainHide
-- Close the main form and deregister events.
------------------------------------------------------------------------------------------------------
function AchievementTracker:MainHide()
 -- Set the config that the main is not visible.
 self.Config.IsVisible = false
 -- Stop timer.
 Apollo.StopTimer("AchievementUpdateTimer")
 -- Unregister event.
 Apollo.RemoveEventHandler("AchievementUpdated", self)
 -- Hide windows.
 self.MainForm:Show(false, false)
 self.OptionsForm:Show(false, false)
end

------------------------------------------------------------------------------------------------------
-- AchievementTracker MainShow
-- Open the main form and do preparations.
------------------------------------------------------------------------------------------------------
function AchievementTracker:MainShow()
 -- Set the config that the main is visible.
 self.Config.IsVisible = true
 -- Update main form.
 self:BuildUpAchievementList()
 -- Register event again and build up the achievement list, because the event is deregisterd on close.
 Apollo.RegisterEventHandler("AchievementUpdated", "OnAchievementUpdated", self)
 self.MainForm:Show(true, false)
 -- Start the update timer.
 Apollo.StartTimer("AchievementUpdateTimer")
end

------------------------------------------------------------------------------------------------------
-- AchievementTracker UpdateAchievementItem
-- Updates all information in the item from the achievement item that is stored in it.
-- Structure of an event item - property - returntype - example
-- item:GetCategory() - string - "Algofels"
-- item:GetCategoryFullName() - string - ""
-- item:GetCategoryId() - number - 197
-- item:GetChecklistItems() - ? - nil
-- item:GetChildTier() - userdata - sub eventitem / nil - todo: this can be used to show some achievments show together
-- item:GetCurrencySystem() - ? - nil
-- item:GetDateCompleted() - string - ""
-- item:GetDescription() - string - "lore ipsum"
-- item:GetGroup() - string - ""
-- item:GetId() - number - 855
-- item:GetName() - string - "Loftit-Rausch"
-- item:GetNumCompleted() - number - 0
-- item:GetNumNeeded() - number - 10
-- item:GetParentTier() - userdate - parent eventitem / nil - todo: this can be used to show some achievments show together
-- item:GetPoints() - number - 10
-- item:GetProgressText() - string - "lore ipsum"
-- item:GetRewards() - table - list of userdata
-- item:GetSubGroup() - string - ""
-- item:GetSubGroupId() - number - 0
-- item:GetTradeskillLayout() - ? - nil
-- item:GetTradeskillRewards() - ? - nil
-- item:GetTradeskillSchematicsRequired() - table - list of userdata
-- item:GetWorldZoneId() - number - 13
-- item:IsChecklist() - boolean - false
-- item:IsComplete() - boolean - false
-- item:IsCurrencyShown() - boolean - false
-- item:IsGuildAchievement() - boolean - false
------------------------------------------------------------------------------------------------------
function AchievementTracker:UpdateAchievementItem(item)
 if item ~= nil then
  local data = item:GetData()
  if data ~= nil then
   -- Animate Update.
   item:SetSprite("CRB_WindowAnimationSprites:sprWinAnim_BirthSmallTempDouble")
   -- Set tooltip for the item self: "AchievementItem".
   item:SetTooltip(self:GetAchievementTooltip(data))
   -- Show "AddToCustomList" only if mode is not custom list.
   item:FindChild("AddToCustomList"):Show(
    self.Config.AchievementTrackerMode ~= ModeCustomListId, true)
   -- Show "RemoveFromCustomList" only if mode is custom list.
   item:FindChild("RemoveFromCustomList"):Show(
    self.Config.AchievementTrackerMode == ModeCustomListId, true)
   -- Show "AchievementProgressBar" only if achievment have amount > 1 and set values.
   local max = data:GetNumNeeded()
   local prefixProgressBar = ""
   local progress = data:GetNumCompleted()
   local progressBar = item:FindChild("AchievementProgressBar")
   -- Set the value to validate if progess is not updated via event.
   progressBar:SetData(progress)
   if max > 1 then
    -- Fix for overfill a achievement.
    if progress > max then
     progress = max
    end
    progressBar:SetMax(max)
    progressBar:SetProgress(progress)
    prefixProgressBar = progress .. "/" .. max .. " "
    if progress > 0 and
       progress < max then
     -- Activate the glow only minimum of one is completed the count is lower as max.
     progressBar:EnableGlow(true)
    else
     progressBar:EnableGlow(false) 
    end
   end
   progressBar:Show(max > 1, true)
   -- Set "AchievementPointsNumber".
   item:FindChild("AchievementPointsNumber"):SetText(data:GetPoints())
   -- Show "AchievementPointsBackerGlow" if achievement is completed.
   item:FindChild("AchievementPointsBackerGlow"):Show(data:IsComplete(), true)
   -- Set icon for "AchievementType" for the category type.
   item:FindChild("AchievementType"):SetSprite(self:GetAchievementCategoryIcon(data:GetCategoryId()))
   -- Show "AchievementComplete" if achievement is completed.
   item:FindChild("AchievementComplete"):Show(data:IsComplete(), true)
   -- Set titel and color for "Titel".
   item:FindChild("Titel"):SetText(
    prefixProgressBar ..
    self:GetAchievementNamePrefix(data) ..
    data:GetName())
   if data:IsComplete() == true then
    item:FindChild("Titel"):SetTextColor("UI_BtnTextGreenNormal")
   else
    item:FindChild("Titel"):SetTextColor("UI_WindowTextDefault")
   end
  end -- data
 end -- item
end

------------------------------------------------------------------------------------------------------
-- ModeListItem OnButtonSignalModeListItem
-- Update the info for the combobox for the selected mode.
------------------------------------------------------------------------------------------------------
function AchievementTracker:UpdateModeComboBox(mode)
 if mode ~= nil then
  local modeComboBox = self.OptionsForm:FindChild("ModeComboBox")
  modeComboBox:SetData(mode)
  modeComboBox:SetText(mode.Name)
  modeComboBox:SetTooltip(mode.Tooltip)
  modeComboBox:SetCheck(false)
 end
end

------------------------------------------------------------------------------------------------------
-- AchievementTracker UpdateTitel
-- Update the titel with current achievementpoints.
------------------------------------------------------------------------------------------------------
function AchievementTracker:UpdateTitel()
 local titel = self.MainForm:FindChild("Titel")
 titel:SetText("Achievement Tracker - " .. AchievementsLib.GetAchievementPoints() .. "P.")
 local tooltip = "Achievement points: " .. AchievementsLib.GetAchievementPoints()
 local recentCompleted = AchievementsLib.GetRecentCompletedAchievements()
 if table.getn(recentCompleted) > 0 then
  tooltip = tooltip .. "\n \nRecently completed achievements:"
  for position, item in pairs(AchievementsLib.GetRecentCompletedAchievements()) do
   tooltip = tooltip .. "\n" .. item:GetName() .. " - " .. item:GetPoints() .. " Points"
  end
 end
 titel:SetTooltip(tooltip)
end

------------------------------------------------------------------------------------------------------
-- AchievementTracker ZoneHasAchievments
-- This function returns true if a given zone id have a achievement otherwise false.
------------------------------------------------------------------------------------------------------
function AchievementTracker:ZoneHasAchievments(zoneId)
 local result = false
 -- Check if the info is already collected.
 if self.ZoneInfo[zoneId] ~= nil then
  result = self.ZoneInfo[zoneId]
 else
  -- Check for personal achievements.
  local achievementList = AchievementsLib.GetAchievements(false)
  for position, item in pairs(achievementList) do
   if item:GetWorldZoneId() == zoneId then
    result = true
    break
   end
  end
  -- Check for guild achievements.
  achievementList = AchievementsLib.GetAchievements(true)
  for position, item in pairs(achievementList) do
   if item:GetWorldZoneId() == zoneId then
    result = true
    break
   end
  end
 end
 return result
end

--****************************************************************************************************
-- AchievementTracker Instance
--****************************************************************************************************
local AchievementTrackerInst = AchievementTracker:new()
AchievementTrackerInst:Init()
